 /***************************************************************************
                          solver.c  -  description
                             -------------------
    begin                : Fri May 12 22:20:46 EDT 2000
    copyright            : (C) 2000 by Andreas Richter
    email                : ar5@rcs.urz.tu-dresden.de
    license              : GNU Public License
 ***************************************************************************/

/*
Copyright (C) 2000  Andreas Richter

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


/* Program to solve the chlorine decay model for different sets of initial
   conditions and model parameters.

   input: initial solution (chlorine concentration at all nodes)
          parameters A0, A1, Peclet number
    number of iterations
    name of output file
    name of tecplot file (visualization)
   output:file with results after the given amount of iterations
          tecplot file

   Compilation
   -----------

   On a Unix/Linux system running the GNU C compiler, the following command can
   be used to compile this file

       gcc -lm -O3 -o solver solver.c

   which produces an executable called solver. Compilation was also tested using
   Microsoft Visual C++ 6.0 on a Microsoft Windows NT 4.0 system.

   Operation
   ---------
   The solver executable is a console-based program carrying out all the
   necessary calculations for solving the partial differential equation
   developed in

   Richter, Andreas: Modeling chlorine decay in dead ends of water distribution
   systems under gerenalized intermmittent flow conditions. Master's thesis,
   University of Cincinnati, Cincinnati, OH, 2001.

   Upon start, the program performs a search for a text configuration file
   called "solver.rc", expected to reside in the same directory as the solver
   executable. The file "solver.rc" is expected to consist of lines having
   the format:

           [parameter]=parameter value

   valid parameters and the expected parameter value are

   [INPUTFILE]      filename of the initial solution
   [OUTPUTFILE]     filename of the solution file
   [TECPLOTFILE]    filename of the output tecplot file
   [CUPMIXFILE]     filename of the cupmixing average storage file
   [ITERATIONS]     number of iterations (integer)
   [A0A]            a value of the  parameter A0,a,
                    inverse of the Peclet number
   [A0R]            a value of the parameter A0r
   [A11M]           a value of the parameter A1,1(location-dependency)
                    (A1,1 = A1,1m*x +A1,1n)
   [A11N]           a value for the parameter A1,1(constant)
   [A12M]           a value for the parameter A1,2(location-dependency)
                    (A1,2 = A1,2m*x +A1,2n)
   [A12N]           a value for the parameter A1,2(constant)
   [A2]             a value of the parameter A2
   [FLOW]           flow regime: "turbulent" for turbulent flow, "laminar"
                    for laminar flow, and any other value for stagnation
   [VELOCITY]       if set to 1, use the velocity profile given in
                    inputfile
   [TIMESTEP]       a value for dtau
   [CUPMIXIV]       intervall after which the cupmixing average should
                    be stored
   [AVERAGE]        "cupmix" means that  the cup-mixing average will be
                    calculated, any other value implies area averaging

   The following two parameters need to be specified, if the inlet chlorine
   concentration changes during the model interval. If left empty,
   a constant inlet concentration of 1 will be assumed.
   [C01]            inlet concentration at the beginning of the model run
   [C02]            inlet concentration at the end of the model run (after
                    n time steps). If left empty, a constant inlet concentration
                    specified by [C01] is assumed.

   The following options are only needed for the calculation of the
   diffusivity  for the turbulent case. A0a and A0r are based on the
   molecular diffusion coefficient (as for laminar flow). Two new constants,
   A0ae and A0re are calculated from the eddy diffusivity at the centerline
   of the pipe. These values are used to calculate the eddy diffusivity at
   every radial position in the pipe. For this purpose, the Reynolds-No. and
   the pipe radius in m are required.
   [RE]             Reynolds-Number
   [R0]             pipe radius (in m)
   [A0AE]           A0A calculated using the eddy diffusivity at the center
                    line of the pipe
   [A0RE]           A0R calculated using the eddy diffusivity at the center
                    line of the pipe

   The Option [MEASUREMENTFILE] can be included in the configuration file to
   specify a file that contains measured chlorine concentrations as a function
   of distance from the pipe inlet. These datasets will be compared to the
   calculated average chlorine concentration profile at the end of the program
   run.
   A plot can be generated to show the measured data compared to the calculated
   results at the end of the model run. The plot file must be specified using
   the option [MEASUREMENTPLOTFILE].
   [MEASUREMENTFILE]     Measurement data file. This file must contain lines
                         with two numbers separated by a tab character. The
                         first number is the distance from the inlet
                         (dimensionless),the second number is the chlorine
                         concentration (dim.-less) at that location.
   [MEASUREMENTPLOTFILE] Measurement plot file.

*/




/* #ifdef HAVE_CONFIG_H
#include <config.h>
#endif */
/* solver.c */
/* This file uses the solution provided by initial.c as initial condition */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
// #include <strings.h>
/* for option - reading */
/* #include <unistd.h> */


/* maximum dimensions of matrices                                          */
/* change this to adapt to computer memory constraints                     */
/* too large values can result in segmentation faults                      */
/* on an Athlon using Linux & 128 MB, the limit was about xmax=ymax=1250   */
/* with model_size=1 */
#define xmax 1000
#define ymax 1000
#define tmax 500

/* defines the number of equations used to solve the model                 */
/* important for blocktridiagonal solver                                   */
/* 1 : tridiagonal matrix                                                  */
/* more than 1: block - tridiagonal matrix                                 */
/* (each element of the matrix consists of a matrix                        */
/* having the "size model_size*model_size                                  */
/* change this value if you want to include more model equations           */
/* increases the size of arrays, so make sure that xmax and ymax have
   reasonable values ( segfault possible )                                 */
#define model_size 1

/* Number of iterations */
long int iterations;
/* Coordinates in physical space */
double xy[xmax][ymax][2];
/* Velocity at each grid point can be stored in this array, if there are
   different velocities throughout the grid                                */
double U[xmax][ymax];
/* coordinate system boundaries (coordinates of the upper right boundary   */
int xi_max, eta_max;
/* concentration arrays; c2 is for temporary storage                        */
double c[xmax][ymax][model_size];
double c2[xmax][ymax][model_size];
/* measurement values, for calculation of model adequacy */
double measure[xmax][2][model_size];
/* cupmixing average */
double cupmix[tmax][xmax];
/* current time level */
double cupmix_compare[xmax];
double t;
int itflag=0;
/* flag for the output (1: output only deviation, 0 output everything ) */
int deviationflag;
/* Model parameters */
double A0ar[xmax][ymax];
double A0rr[xmax][ymax];
double A0a, A0r, A11m, A12m, A11n, A12n, A2;
double r, x, d_x, d_x1, d_x2, d_r1,d_r2, d_r, d_t;
int cupmix_intervall;
/* for eddy diffusivity calculation */
double Re, r0, D0, u0, A0ae, A0re;
double viscosity = 1e-6; // units: [m2s]
double DD[xmax][ymax];
/* for calculation of varying inlet concentration */
double c01, c02, dc0;
/* flow regime description */
int flow=0;
/* which averaging value to use (cupmix or other) */
int average=0;
/* flag for velocity profile description */
int velocityprofile=0;
/* filenames */
char *logfilename;
char *inputfilename;
char *outputfilename;
char *configfilename;
char *tecplotfilename;
char *cupmixfilename;
char *measurementfilename;
char *measureplotfilename;
char *plotlabel;
char *multiplerunsplotfilename;
char *locationwithtimeplotfilename;
/* FILE pointers */
FILE *logfile;
FILE *cupmixfile;
/* matrices for solution scheme */
double A[ymax][model_size][model_size];
double B[ymax][model_size][model_size];
double C[ymax][model_size][model_size];
double RHS[ymax][model_size];
/* flags for checking the input */
int is_outputfile, is_inputfile;
int is_tecplotfile, is_cupmix;
int is_measure,is_multiplerunsplot;
int is_locationplot;
/* if this is a run in a series: what running number to use in the plot */
int runNumber;
/* save average how often ?*/
int times;
int currentaverageNo;
/* how many measurement datasets do we have*/
int mpoints;
/* info for printing the results at the important locations */
/* Xcup contains position (as fraction of total length */
/* Ccup will contain Concentration at these locations */
double Xcup[10]={9.41e-4, 5.35e-2, 0.1958, 0.3381, 0.4487, 0.5838, 0.6943, 0.8367, 0.97900,1};
double Ccup[10];
/* info for printing cupmixing average at desired locations */
double Xcup_user[20];
int Nr_cup_user=0;
/* location for plot of chlorine concentration at one location, with time */
double location;
/* time, for plot of chlorine concentration at one location */
double locPlot_time;

void block_tridiagonal(int IU,double A[][model_size][model_size],
                              double B[][model_size][model_size],
                              double C[][model_size][model_size],
                              double D[][model_size])
     /* solves a block tridiagonal matrix system. Matrix elements
        (blocks) can be of size
  up to 4x4, tridiagonal matrices (1x1) are also possible.
  IU contains the length of the matrix(not the length of the blocks)
  A contains the elements of the matrix left of the main diagonal
  B contains the elements of the main diagonal
  C contains the elements of the matrix left of the main diagonal
  D contains the right hand side (RHS) of the equation. It will also
  contain the solution vector */
{
  int k;
  void LU_deco(double matr[][model_size]);
  void LU_solv1(double matr[][model_size],double vect[]);
  void LU_solv2(double matr1[][model_size],double matr2[][model_size]);
  void Multi1(double vect1[], double matr[][model_size],double vect2[]);
  void Multi2(double matr1[][model_size],double matr2[][model_size],
              double matr3[][model_size]);
  LU_deco(B[0]);
  LU_solv1(B[0],D[0]);
  LU_solv2(B[0],C[0]);
  for(k=1;k<IU;k++)
  {
    Multi1(D[k],A[k],D[k-1]);
    Multi2(B[k],A[k],C[k-1]);
    LU_deco(B[k]);
    LU_solv1(B[k],D[k]);
    LU_solv2(B[k],C[k]);
  }
  for(k=IU-2;k>=0;k--)
  {
    Multi1(D[k],C[k],D[k+1]);
  }
}

void LU_deco(double matr[][model_size])
{
  int k,m,i,j;
  for(k=0;k<=model_size-2;k++)
  {
    for(m=k+1;m<=model_size-1;m++)
      matr[m][k]=matr[m][k]/matr[k][k];
    for(i=k+1;i<=model_size-1;i++)
      for(j=k+1;j<=model_size-1;j++)
        matr[i][j]=matr[i][j]-matr[i][k]*matr[k][j];
  }
}
void LU_solv1(double matr[][model_size],double vect[])
{
  int i,j;
  double sum;

  for(i=1;i<=model_size-1;i++)
  {
    sum=0.0;
    for(j=0;j<=i-1;j++)
      sum=sum+matr[i][j]*vect[j];
    vect[i]=vect[i]-sum;
  }
  vect[model_size-1]=vect[model_size-1]
    /matr[model_size-1][model_size-1];
  for(i=model_size-2;i>=0;i--)
  {
    sum=0;
    for(j=i+1;j<=model_size-1;j++)
      sum=sum+matr[i][j]*vect[j];
    vect[i]=(vect[i]-sum)/matr[i][i];
  }
}

void LU_solv2(double matr1[][model_size],double matr2[][model_size])
{
  int i,j,k;
  double sum;
  for(k=0;k<=model_size-1;k++)
  {
    for(i=1;i<=model_size-1;i++)
    {
      sum=0.0;
      for(j=0;j<=i-1;j++)
        sum=sum+matr1[i][j]*matr2[j][k];
      matr2[i][k]=matr2[i][k]-sum;
    }
    matr2[model_size-1][k]=matr2[model_size-1][k]
      /matr1[model_size-1][model_size-1];
    for(i=model_size-2;i>=0;i--)
    {
      sum=0;
      for(j=i+1;j<=model_size-1;j++)
        sum=sum+matr1[i][j]*matr2[j][k];
      matr2[i][k]=(matr2[i][k]-sum)/matr1[i][i];
    }
  }
}


void Multi1(double vect1[], double matr[][model_size],double vect2[])
{
  int i,j;
  double sum;
  for(i=0;i<=model_size-1;i++)
  {
    sum=0;
    for(j=0;j<=model_size-1;j++)
      sum=sum+matr[i][j]*vect2[j];
    vect1[i]=vect1[i]-sum;
  }
}

void Multi2(double matr1[][model_size],double matr2[][model_size],
            double matr3[][model_size])
{
  int i,j,k;
  double sum;
  for(k=0;k<=model_size-1;k++)
    for(i=0;i<=model_size-1;i++)
    {
      sum=0;
      for(j=0;j<=model_size-1;j++)
        sum=sum+matr2[i][j]*matr3[j][k];
      matr1[i][k]=matr1[i][k]-sum;
    }
}

/* READINPUTFILE - opens the file specified by inputfilename and reads the
   data. Dimensions of the array are specified in the first two lines of the
   file. */

void readinputfile()
{
  int i,j;
  char xs[15];
  char ys[15];
  char v1s[15];
  char v2s[15];
  char v3s[15];
  FILE *datafile;
  char s[20];
  datafile=fopen(inputfilename,"r");
  if (datafile==NULL) {
    /* opening was unsuccessful */
    fprintf(stderr,"Couldn't open %s for read access.\n", inputfilename);
    exit(8);
  }
  else {
    // read dimensions of array
    fscanf(datafile,"Xmax = %d\n",&xi_max);
    fscanf(datafile,"Rmax = %d\n",&eta_max);
    if ((xi_max>xmax)||(eta_max>ymax)) {
      // we do not have enough memory available:
      fprintf(stderr,
        "Array size definition in the code is to small to be able to\n");
      fprintf(stderr,"load the input file %s.\n",inputfilename);
      fprintf(stderr,
        "Edit the code (change xmax, ymax), or use a smaller grid.\n");
      fclose(datafile);
      exit(8);
    } else {
      // read dimensionalization info (info is not used by code)
      fscanf(datafile,"%s%s%s%s%s\n",s,s,s,s,s);
      // read array
      for (i=0; i<=xi_max-1; i++) {
        for (j=1; j<=eta_max; j++) {
          fscanf(datafile,"%s %s %s %s %s\n",xs, ys, v1s, v2s, v3s);
          xy[i][j][0]=atof(xs);
          xy[i][j][1]=atof(ys);
          c[i][j][0]=atof(v1s);
          if (model_size>1) c[i][j][1]=atof(v2s);
          if (velocityprofile) U[i][j]=atof(v3s);
          else U[i][j]=1;
        }
        // symmetry at the lower boundary
        // copy the values at j=1 to matrix position j=0
        c[i][0][0]=c[i][1][0]; c[i][0][1]=c[i][1][1];
        U[i][0]=U[i][1];
        xy[i][0][0]=xy[i][1][0];
        xy[i][0][1]=-xy[i][1][1];
      }
      fclose(datafile);
      // inflow boundary (i=0)
      for (j=0;j<=eta_max;j++) c[0][j][0]=c01;
    }
  }
}

/* READMEASUREMENTFILE - opens a file containing measurement data, which is
expected to be listed one dataset (x,c) per line, the two values per line
are separated by tab characters. Data will be stored in cm for further analysis
 */

void readmeasurementfile()
{
  FILE *datafile;
  int i,check;
  int result;
  char xs[20];
  char ys[20];
  datafile=fopen(measurementfilename,"r");
  if (datafile==NULL) {
    /* opening was unsuccessful */
    fprintf(stderr,"Couldn't open %s for read access.\n", inputfilename);
    exit(8);
  }
  else {
    i=0;
    check=0;
    while (check==0) {
      result=fscanf(datafile,"%s\t%s\n",xs,ys);
      if (result==2) {
        measure[i][0][0]=atof(xs);
        measure[i][1][0]=atof(ys);
        printf("%d:  %e  %e\n",i+1,measure[i][0][0],measure[i][1][0],xs,ys);
        i++;
      }
      else {
        check=1;
        if (result==EOF) printf ("End of file. %d measurement datasets successfully loaded.\n",i);
        else {
          printf("Unexpected value encountered. Stopping read.\n");
          is_measure=0;
        }
      }
    }
    mpoints=i;
  }
}

void extract_data()
{
  int i;
  int j;
  int n;
  double dx_measurement;
  double dx_code;
  double df_measurement;
  double df_code;
  n=times;
  j = 0;
  for (i=0;i<mpoints;i++){
    while (xy[j][0][0]<measure[i][0][0]){
      j++;
    }
    df_code=cupmix[n][j]-cupmix[n][j-1];
    dx_code=xy[j][0][0]-xy[j-1][0][0];
    dx_measurement = xy[j][0][0]-measure[i][0][0];
    //fvalues[i][0][0]=measure[i][0][0];
    cupmix_compare[i]=cupmix[n][j-1]+df_code*dx_measurement/dx_code;
    printf("%d: %f %f    %f %f\n",i,measure[i][0][0],measure[i][1][0],xy[j-1][0][0],xy[j][0][0]);
    //printf("%d %d: %lf  %lf %lf %lf\n",i,j, [i][0][0],[i][1], cupmix[0][j], cupmix[0][j-1]);
  }
}

double f(double r)
{
  double rr;
  /* return a value of f depending on the radius and the flow pofile */
  /* flow =1 : turbulent, flow = 0 : laminar , flow=2: stagnant*/
  switch(flow) {
    case 1 :  rr=1; break;
    case 0 :  rr=2*(1-r*r); break;
    case 2 :  rr=0;
  }
  return rr;
}

/* OPENCUPMIXFILEMTV
   creates a file that will be used to store the cupmixing concentration
   as an plotmtv-readable file
*/
void open_cupmixfile_mtv()
{
  cupmixfile=fopen(cupmixfilename,"w");
  if (cupmixfile==NULL) {
    /* opening was unsuccessful */
    fprintf(stderr,"Couldn't open %s for write access .\n", cupmixfilename);
    exit(8);
  }
  else {
    times = (int)((double)iterations/(double)cupmix_intervall)+1;
    if (times>tmax) {
      fprintf(stderr,"Too many intervalls for saving the cupmixing average.\n");
      fprintf(stderr,"Please change the values for [CUPMIXIV] or [ITERATIONS]\n");
      fprintf(stderr,"or modify the code (increase tmax) to fit your needs.\n");
      fclose (cupmixfile);
      exit(8);
    } else {
      printf ("Saving cupmixing average at %d intervalls (%d/%d+1).\n",
        times,iterations,cupmix_intervall);
      fprintf(cupmixfile, "$ DATA=CURVE2D\n");
      fprintf(cupmixfile, "%% toplabel = \"Cup-Mixing Average Chlorine");
      fprintf(cupmixfile, " Concentration with time\"\n");
      fprintf(cupmixfile, "%% subtitle = \"A0a=%.2f, A_0r=%.2f, ",A0a,A0r);
      if (flow!=1) {
        fprintf(cupmixfile, "R=");
        if (A11m!=0) {
          fprintf(cupmixfile, "(%.3f*X+%.3f)C",A11m,A11n);
        } else {
          fprintf(cupmixfile, "%.3f*C",A11n);
        }
        if (A12m!=0) {
          fprintf(cupmixfile, "+(%.3f*X+%.3f)C, ",A12m,A12n);
        } else {
          fprintf(cupmixfile, "+%.3f*C, ",A12n);
        }
      }
      else {
      }
      fprintf(cupmixfile, "A2=%.3f\"\n",A2);
      fprintf(cupmixfile, "%% xlabel = \"X\"\n");
      fprintf(cupmixfile, "%% ylabel = \"C\"\n");
      fprintf(cupmixfile, "%% ymin=0 ymax=1.1\n");
    }
  }
}


void close_cupmixfile_mtv()
{
  fprintf(cupmixfile,"$ END\n");
  fclose(cupmixfile);
}

/* CUPMIXINGaverageMTV - calculate the average concentration at each axial
   distance Calculation method (cup-mixing or area-averaging) can be chosen in
   solver.rc file
                                                                           */
void cupmixing_average_mtv(int n)
{
  int i,j,k;
  double dr, r;
  double rmax;
  double sum;
  printf("Saving cupmixing average at t = %f, outlet:",t);
  fprintf(cupmixfile,"%% linelabel=\"T = %f\"\n",t);
  fprintf(cupmixfile,"%% linecolor=%d\n",0);
  fprintf(cupmixfile,"%% linetype =%d\n",n);
  //fprintf(cupmixfile,"%% type =\"n = %d\"\n",n);
  for (i=0;i<xi_max;i++) {
    sum = 0;
    for (j=1;j<eta_max;j++){
      r = (xy[i][j][1]+xy[i][j+1][1])/2;
      dr = (xy[i][j+1][1]-xy[i][j][1]);
      if (average==1) { // cup-mixing average
        if (flow==1) { // turbulent or stagnant
          sum+=2*(c[i][j][0]+c[i][j+1][0])*0.5*r*dr;
        }
        else { // laminar or stagnant
          sum+=4*(c[i][j][0]+c[i][j+1][0])*0.5*(r-r*r*r)*dr;
        }
      }
      else /* cross-sectional average */ sum+=c[i][j][0]*(2*r*dr);
    }
    cupmix[n][i]=sum;
    for(k=9;k>=0;k--) {
      if ((xy[i][1][0]>=Xcup[k])&&(xy[i-1][1][0]<Xcup[k])) Ccup[k]=sum;
    }
    /*if ((xy[i][1][0]>=X3)&&(xy[i-1][1][0]<X3)) C3=sum;
    else if ((xy[i][1][0]>=X2)&&(xy[i-1][1][0]<X2)) C2=sum;
      else if ((xy[i+1][1][0]>X1)&&(xy[i][1][0]<=X1)) C1=sum;*/
    fprintf(cupmixfile," %f %f\n",xy[i][1][0],cupmix[n][i]);
  }
  fprintf(cupmixfile,"\n");
  printf("Cout = %e\n",cupmix[n-1][xi_max-1]);
  printf("Concentrations in between: 0: %e 5: %e 9: %e\n",Ccup[0],Ccup[5],Ccup[9]);
}


/* CUPMIXINGAVERAGE - calculate the average concentration at each axial
   distance. Calculation method (cup-mixing or area-averaging) can be
   chosen in .rc file                                                      */
void cupmixingaverage()
{
  int i,j;
  double dr, r;
  double rmax;
  double sum;
  printf("Saving cupmixing average at t = %f, outlet:",t);
  if (t<=0) fprintf(cupmixfile,"Here\n");
  for (i=0;i<xi_max;i++) {
    sum = 0;
    for (j=1;j<eta_max;j++){
      r = (xy[i][j][1]+xy[i][j+1][1])/2;
      dr = (xy[i][j+1][1]-xy[i][j][1]);
      if (average==1) { // cup-mixing averagefloat
        if (flow==1) { // turbulent or stagnant
          sum+=2*(c[i][j][0]+c[i][j+1][0])*0.5*r*dr;
        }
        else { // laminar or stagnant
          sum+=4*(c[i][j][0]+c[i][j+1][0])*0.5*(r-r*r*r)*dr;
        }
      }
      else /* cross-sectional average */ sum+=c[i][j][0]*(2*r*dr);
    }
    cupmix[0][i]=sum;
    /*if ((xy[i][1][0]>=X3)&&(xy[i-1][1][0]<X3)) C3=sum;
    else if ((xy[i][1][0]>=X2)&&(xy[i-1][1][0]<X2)) C2=sum;
      else if ((xy[i+1][1][0]>X1)&&(xy[i][1][0]<=X1)) C1=sum;*/
    fprintf(cupmixfile,"%f %f %f\n",t,xy[i][1][0],cupmix[0][i]);
  }
  printf("Cout = %e\n",cupmix[0][xi_max-1]);
  //printf("Concentrations in between: 0: %e 5: %e 9: %e\n",C1,C2,C3);
}

/* REACTIONTERM_RUNGEKUTTA - calculates the reactionterm R based on a 4th
   order Runge-Kutta method                                                */
double ReactionTerm_RungeKutta(double C, double dt, double t, double x)
{
  double dcdt,dcdt0, dcdt1, dcdt2, dcdt3;
  double A11, A12;
  A11=A11m*x+A11n;
  A12=A12m*x+A12n;
  dcdt0=(-1)*A11*C-A12*C*C;
  dcdt1=-A11*(C+0.5*dt*dcdt0)-A12*(C+0.5*dt*dcdt0)*(C+0.5*dt*dcdt0);
  dcdt2=-A11*(C+0.5*dt*dcdt1)-A12*(C+0.5*dt*dcdt1)*(C+0.5*dt*dcdt1);
  dcdt3=-A11*(C+dt*dcdt2)-A12*(C+dt*dcdt2)*(C+dt*dcdt2);
  dcdt=dcdt0/6+dcdt1/3+dcdt2/3+dcdt3/6;
  return(dcdt);
}

/* CALCULATEA0 - calculates the constants A0A and A0R based on the location
   in the pipe. Laminar: constant value, Turbulent: use Law of the pipe    */
void calculateA0(double r, double *A0ar, double *A0rr, int i, int j)
{
  double e = 2.59e-4; // pipe wall roughness [m]
  double f;           //
  double K=0.4;       // Karman constant
  double dummy;
  double D;
  double vt;
  int useconstant=0;
  // laminar or stagnant case, no calculation neccessary.
  if ((flow==0)||(flow==2)||(useconstant==1)) {
    *A0ar=A0a;
    *A0rr=A0r;
  }
  // turbulent
  else {
    /* Moody chart equation */
    dummy = log(e/3.7/2/(r0)+5.74/pow(Re,0.9));
    f=1.32547/dummy/dummy;
    /* Reichardt's law of the pipe (Edwards) */
    vt = sqrt(f/8)*(u0*U[i][j]);
    D=K*r0*(1-r)*vt*(1+r)*(1+2*r*r)/6;
    if (r==0) {
      D0 = D;
      *A0ar=A0a+A0ae;
      *A0rr=A0r+A0re;
    }
    else if (D0>0) {
      *A0ar=A0a+D/D0*A0ae;
      *A0rr=A0r+D/D0*A0re;
    }
    // save the diffusion coefficient
    DD[i][j]=*A0rr;
  }
}
/* SOLVE : main fuction to solve the system of linear equations.
           Loads the matrix and inverts it
*/

void artificial_diffusion(int i)
{
  int j;
  //printf("Diffusion at %d\n",i);
  for (j=2;j<eta_max;j++)
    c[i][j][0]=(c2[i][j][0]*4+c2[i+1][j][0]
                             +c2[i-1][j][0]
                             +c2[i][j+1][0]
                             +c2[i][j-1][0])/8;
  for (j=2;j<eta_max;j++)
    c2[i][j][0]=c[i][j][0];

}

void solve()
{
  int n,i,j;
  int offset;
  int printerror=1;
  int runcounter=1;
  double fr;
  double cmix=0;
  double reactionterm;
  double A0adummy, A0rdummy;
  /* initiate eddy diffusivity calculation*/
  calculateA0(0,&A0adummy,&A0rdummy,0,0);
  for (i=0;i<xi_max;i++){
    for (j=0;j<=eta_max;j++) {
      r=xy[i][j][1];
      calculateA0(r,&A0adummy,&A0rdummy,i,j);
      A0ar[i][j]=A0adummy;
      A0rr[i][j]=A0rdummy;
    }
  }
  /* matrix handling */
  offset=0;
  /* number of iterations, start iteration */
  t=0;
  //  save first initial distribution of average concentration
  if (is_cupmix) cupmixing_average_mtv(1);
  currentaverageNo=1;
  /************************************************************************ */
  for (n=1;n<=iterations;n++){
    /* first half step (n+1/2), differenciating in r direction */
    /* we start counting in x direction */
    printerror=1;
    for(i=1;i<xi_max;i++) {
      /* r direction */
      for(j=1;j<=eta_max;j++) {
        /* calculate r, f(r) */
        r=xy[i][j][1];
        fr=f(r);
        /* calculate reaction term */
        reactionterm=ReactionTerm_RungeKutta(c[i][j][0],d_t/2,t,xy[i][j][0]);
        /* create right hand side */
        if (i<xi_max-1) {
          // any point except for the outflow boundary
          d_x=(xy[i+1][j][0]-xy[i-1][j][0])/2;
          d_x1=xy[i][j][0]-xy[i-1][j][0];
          d_x2=xy[i+1][j][0]-xy[i][j][0];
          RHS[offset+j-1][0]= c[i][j][0]
              +d_t/2*A0ar[i][j]/d_x
                 *((c[i+1][j][0]-c[i][j][0])/d_x2
                  -(c[i][j][0]-c[i-1][j][0])/d_x1)
              -d_t*0.25/d_x*fr
                 *((c[i+1][j][0]-c[i-1][j][0])*U[i][j]
                  +(U[i+1][j]-U[i-1][j])*c[i][j][0])
              +reactionterm*d_t*0.5
              +d_t*0.25/d_x*(U[i+1][j]-U[i-1][j])*c[i][j][0]*fr;
        } else {
          // outflow boundary
          d_x=(xy[i][j][0]-xy[i-1][j][0]);
          RHS[offset+j-1][0] =
              c[i][j][0]
              -d_t*0.5/d_x*fr
                *((c[i][j][0]-c[i-1][j][0])*U[i][j]
                 +(U[i][j]-U[i-1][j])*c[i][j][0])
              +d_t*0.5*reactionterm
              +d_t*0.5/d_x*(U[i][j]-U[i-1][j])*c[i][j][0]*fr;
        }
        //DD[i][j]=RHS[offset+j-1][0];
      }
      for(j=2;j<eta_max;j++) {
        d_r=(xy[i][j+1][1]-xy[i][j-1][1])/2;
        d_r1=(xy[i][j][1]-xy[i][j-1][1]);
        d_r2=(xy[i][j+1][1]-xy[i][j][1]);
        r=xy[i][j][1];
        fr=f(r);
        /* create matrix (A, B and C)*/
        A[offset+j-1][0][0]=A0rr[i][j]/r*0.25*d_t/d_r
                           -A0rr[i][j]*0.5*d_t/d_r/d_r1
                           +(A0rr[i][j+1]-A0rr[i][j-1])*0.125/d_r/d_r*d_t;
        B[offset+j-1][0][0]=1+A0rr[i][j]*d_t/d_r1/d_r2;
        C[offset+j-1][0][0]=A0rr[i][j]*-0.25/r*d_t/d_r
                           -A0rr[i][j]*0.5*d_t/d_r2/d_r
                           -(A0rr[i][j+1]-A0rr[i][j-1])*0.125/d_r/d_r*d_t;
      }
      /* ***************************** boundaries **************************/
      /* symmetry boundary condition */
      r=xy[i][1][1]; fr=f(r);
      d_r=(xy[i][2][1]-xy[i][1][1]);
      B[offset][0][0]=1+A0rr[i][1]/d_r/d_r*d_t*0.5;
      C[offset][0][0]=-A0rr[i][1]/d_r/d_r*d_t*0.5;
      /* wall boundary condition: */
      r=xy[i][eta_max][1]; fr=f(r);
      d_r=(xy[i][eta_max][1]-xy[i][eta_max-1][1]);
      A[offset+eta_max-1][0][0]=(-1)*A0rr[i][eta_max]/d_r/d_r*d_t*0.5;
      B[offset+eta_max-1][0][0]=1+((A0rr[i][eta_max]-A0rr[i][eta_max-1])*A2/d_r
                                   +A0rr[i][eta_max]/r*A2
                                   +A2*A0rr[i][eta_max]/d_r
                                   +A0rr[i][eta_max]/d_r/d_r
                                   )*d_t*0.5;
      /* execute solver */
      block_tridiagonal(offset+eta_max,A,B,C,RHS);
      /* decode solution vector (RHS) */
      for(j=1;j<=eta_max;j++) {
        c2[i][j][0]=RHS[offset+j-1][0];
        if (c2[i][j][0]<0) {
          //c2[i][j][0]=0;
          if (printerror==1) {
            printf("concentration <0 at %d,%d (1st half step). Fixing.",i,j);
            printerror=0;
            printf("t = %f\n",t);
          }
        }
        else if (c2[i][j][0]>1.5) {
            c2[i][j][0]=1.5;
            if (printerror==1) {
              printf("concentration >1.5 at %d,%d (1st half step). Fixing.",
                i,j);
              printerror=0;
              printf("t = %f\n",t);
            }
        }
      }
      /* copy to ghost point - symmetry boundary condition */
      c2[i][0][0]=c2[i][1][0];
      /* add artificial diffusion */
    } //end for (i...xi_max)
    for(i=2;i<xi_max-1;i++)
      if (U[i][0]!=U[i-1][0]) artificial_diffusion(i);
    /* copy solution back to original array */
    for (i=1;i<=xi_max-1;i++) for (j=0;j<=eta_max;j++) {
        c[i][j][0]=c2[i][j][0];
    }
    // cupmixing_average_mtv(2);
    /* change inlet concentration */
    for (j=0;j<=eta_max;j++) {
      c[0][j][0]+=dc0;
    }
    /* ***************************** 2 nd half step ********************** */
    printerror=1;
    for (j=1;j<=eta_max;j++){
      /* calculate r, f(r) (only once for each r) */
      r=xy[0][j][1];
      fr=f(r);
      /* calculate A0rr, A0ar */
      // calculateA0(r, &A0ar, &A0rr,i,j);
      /* x direction */
      for (i=1;i<xi_max;i++){
        /* calculate reaction term */
        reactionterm=ReactionTerm_RungeKutta(c[i][j][0],d_t*0.5,t,
                      xy[i][j][0]);
        /* for all points except the wall boundary */
        if ((j<eta_max)&&(j>1)) {
          // create dr
          d_r=(xy[i][j+1][1]-xy[i][j-1][1])/2;
          d_r1=xy[i][j][1]-xy[i][j-1][1];
          d_r2=xy[i][j+1][1]-xy[i][j][1];
          /* create right hand side */
          RHS[i-1][0]=c[i][j][0]*(1-A0rr[i][j]*d_t/d_r1/d_r2)
                     +reactionterm*0.5*d_t
                     +c[i][j-1][0]*d_t*
                       (A0rr[i][j]/2/d_r1/d_r-A0rr[i][j]/4/r/d_r
                        -(A0rr[i][j+1]-A0rr[i][j-1])/8/d_r/d_r)
                     +c[i][j+1][0]*d_t*
                       (A0rr[i][j]/4/r/d_r+A0rr[i][j]/2/d_r2/d_r
                        +(A0rr[i][j+1]-A0rr[i][j-1])/8/d_r/d_r);
        } else if (j==eta_max) {
            /* wall boundary (j=eta_max)*/
            d_r=xy[i][eta_max][1]-xy[i][eta_max-1][1];
            /* RHS */
            RHS[i-1][0]=c[i][eta_max][0]*
                          (1+(-(A0rr[i][eta_max]-A0rr[i][eta_max-1])/d_r*A2
                              -A0rr[i][eta_max]/r*A2
                              -A2*A0rr[i][eta_max]/d_r
                              -A0rr[i][eta_max]/d_r/d_r)*d_t*0.5)
                        +c[i][eta_max-1][0]*
                         (A0rr[i][eta_max]/d_r/d_r)*d_t*0.5
                        +reactionterm*d_t*0.5;
          } else {
            // symmetry boundary condition (j=1)
            d_r=(xy[i][2][1]-xy[i][1][1]);
            RHS[i-1][0]=A0rr[i][1]/d_r*(c[i][2][0]-c[i][1][0])/d_r*d_t*0.5
                       +reactionterm*d_t*0.5
                       +c[i][1][0];
        }
        if (i<xi_max-1) {
          // not at the outflow boundary
          d_x=(xy[i+1][j][0]-xy[i-1][j][0])/2;
          d_x1=xy[i][j][0]-xy[i-1][j][0];
          d_x2=xy[i+1][j][0]-xy[i][j][0];
          /* A, B, C */
          A[i-1][0][0]=(-1)*A0ar[i][j]/2*d_t/d_x1/d_x-fr*0.5*d_t/d_x*0.5*U[i][j];
          B[i-1][0][0]=1 + A0ar[i][j]*d_t*0.5*(1/d_x1/d_x+1/d_x2/d_x)
                         + fr*d_t*0.25/d_x*(U[i+1][j]-U[i-1][j])
                         - fr*d_t*0.25/d_x*(U[i+1][j]-U[i-1][j]);
          C[i-1][0][0]=(-1)*A0ar[i][j]/2*d_t/d_x2/d_x+fr*0.5*d_t/d_x*0.5*U[i][j];
        } else {           /* outflow boundary */
          d_x=(xy[i][j][0]-xy[i-1][j][0]);
          A[i-1][0][0]=(-1)*fr*0.5*d_t/d_x*U[i][j];
          B[i-1][0][0]=1+fr*0.5*d_t/d_x*(2*U[i][j]-U[i-1][j])
                        -fr*0.5*d_t/d_x*(U[i][j]-U[i-1][j]);
          C[i-1][0][0]=0;
        }
      }
      /* boundaries :*/
      DD[i][j]=A[i-1][0][0]+B[i-1][0][0]+C[i-1][0][0];
      // inflow boundary condition */
      RHS[0][0]-=A[0][0][0]*c[0][j][0];
      /* execute solver */
      block_tridiagonal(xi_max-1,A,B,C,RHS);
      /* decode solution vector */
      for (i=1;i<=xi_max-1;i++) c2[i][j][0]=RHS[i-1][0];
    }
    /****************************** */
    /* copy solution back to original matrix */
    for (j=1;j<=eta_max;j++){
      for (i=1;i<=xi_max-1;i++) {
         if (c2[i][j][0]<0) {
           //c[i][j][0]=0;
           if (printerror==1) {
             printf("concentration <0 @ %d, %d. 2nd half-step. Fixing.",i,j);
             printf("t = %f\n",t);
             printerror=0;
           }
         } else if(c2[i][j][0]>1.5) {
               c[i][j][0]=1.5;
               if (printerror==1) {
                 printf("concentration >1.5 @ %d, %d. 2nd ");
                 printf("half-step.Fixing.\n",i,j);
                 printf("t = %f\n",t);
                 printerror=0;
               }
           } else c[i][j][0]=c2[i][j][0];
      }
    }
    /* copy solution to ghost point */
    for(i=1;i<=xi_max-1;i++) c[i][0][0]=c[i][1][0];
    /* prepare next loop, */
    cmix+=d_t;
    t+=d_t;
    /* save cupmixing average if we have to*/
    //printf("%e %e",cmix,cupmix_intervall);
    if ((runcounter==cupmix_intervall)&&(is_cupmix)) {
      currentaverageNo++;
      //printf("--> saving");
      //cupmixing_average_mtv(1+(int)(t/cupmix_intervall));
      cupmixing_average_mtv(currentaverageNo);
      cmix=0;
      runcounter=0;
    }
    runcounter++;
    //else printf("\n");
  }
}

/* SAVEFILE - writes the result of the calculation into a file that has the
   same structure as the input file.                                       */


void savefile()
{
  int i,j;
  FILE *outputgridfile;
  outputgridfile = fopen(outputfilename,"w");
  if (outputgridfile==NULL) {
    /* opening was unsuccessful */
    fprintf(stderr,"Couldn't open %s for write access.\n", outputfilename);
    exit(8);
  }
  else {
    /* successful opening of the file for write access */
    fprintf(outputgridfile,"Xmax = %d\n",xi_max);
    fprintf(outputgridfile,"Rmax = %d\n",eta_max);
    fprintf(outputgridfile,"x  y  c  X  U\n");
    for (i=0; i<=xi_max-1; i++) {
      for (j=1; j<=eta_max; j++){
          fprintf(outputgridfile,"%e %e %e %e %e\n",xy[i][j][0],xy[i][j][1],
          c[i][j][0],DD[i][j],U[i][j]);
      }
    }
    fclose(outputgridfile);
  }
}

/* OUTPUTMTV - create an output file suitable for
   plotmtv */
void outputmtv()
{
  int i,j;
  FILE *outputfile;
  outputfile = fopen(tecplotfilename,"w");
  if (outputfile==NULL) {
    /* opening was unsuccessful */
    fprintf(stderr,"Couldn't open %s for write access.\n", tecplotfilename);
    exit(8);
  }
  else {
    /* successful opening of the file for write access */
    fprintf(outputfile,"$ DATA=CONTOUR\n");
    fprintf(outputfile,"%% contstyle=2\n");
    // fprintf(outputfile,"%% meshplot\n");
    fprintf(outputfile,"%% xlabel=\"X\"\n");
    fprintf(outputfile,"%% ylabel=\"r\"\n");
    fprintf(outputfile,"%% toplabel=\"Chlorine concentration profile\"\n");
    fprintf(outputfile, "%% subtitle = \"A0a=%.2e, A_0r=%.2e, ",A0a,A0r);
    if (flow!=1) {
      fprintf(outputfile, "R=");
      if (A11m!=0) {
        fprintf(outputfile, "(%.3e*X+%.3e)C",A11m,A11n);
      } else {
        fprintf(outputfile, "%.3e*C",A11n);
      }
      if (A12m!=0) {
        fprintf(outputfile, "+(%.3e*X+%.3e)C, ",A12m,A12n);
      } else {
        fprintf(outputfile, "+%.3e*C, ",A12n);
      }
    }
    else {
    }
    fprintf(outputfile, "A2=%.3e\"\n",A2);
    fprintf(outputfile,"%% equalscale=false\n");
    fprintf(outputfile,"%% fitpage=false\n");
    fprintf(outputfile,"%% xyratio=0.3\n");
    fprintf(outputfile,"%% nx=%d XGRID\n",xi_max);
    for (i=0;i<=xi_max-1;i++)
      fprintf(outputfile," %.3e",xy[i][0][0]);
    fprintf(outputfile,"\n");
    fprintf(outputfile,"%% ny=%d YGRID\n",eta_max);
    for (j=1;j<=eta_max;j++)
      fprintf(outputfile," %.3e",xy[0][j][1]);
    fprintf(outputfile,"\n\n");

    for (j=1; j<=eta_max; j++){
      for (i=0; i<=xi_max-1; i++) {
        fprintf(outputfile," %.5e\n",c[i][j][0]);
        //fprintf(outputfile," %.5f\n",DD[i][j]);
      }
    }

    fprintf(outputfile,"$ END\n");
    fclose(outputfile);
  }
}

/* create two line plots containing the measurement data and the last cupmixing
data */

void measureplot()
{
int i,j;
  FILE *outputfile;
  outputfile = fopen(measureplotfilename,"w");
  if (outputfile==NULL) {
    /* opening was unsuccessful */
    fprintf(stderr,"Couldn't open %s for write access.\n", measureplotfilename);
    exit(8);
  }
  else {
    /* successful opening of the file for write access */
    fprintf(outputfile,"$ DATA=CURVE2D\n");
    //fprintf(outputfile,"%% contstyle=2\n");
    // fprintf(outputfile,"%% meshplot\n");
    fprintf(outputfile,"%% xlabel=\"X\"\n");
    fprintf(outputfile,"%% ylabel=\"C\"\n");
    fprintf(outputfile,"%% toplabel=\"Chlorine concentration profile (average)\"\n");
    fprintf(outputfile,"%% subtitle = \"A0a=%.2e, A_0r=%.2e, ",A0a,A0r);
      if (flow!=1) {
        fprintf(outputfile, "R=");
        if (A11m!=0) {
          fprintf(outputfile, "(%.3e*X+%.3e)C",A11m,A11n);
        } else {
          fprintf(outputfile, "%.3e*C",A11n);
        }
        if (A12m!=0) {
          fprintf(outputfile, "+(%.3e*X+%.3e)C, ",A12m,A12n);
        } else {
          fprintf(outputfile, "+%.3e*C, ",A12n);
        }
      }
      else {
      }
      fprintf(outputfile, "A2=%.3e\"\n",A2);
    //fprintf(outputfile,"%% equalscale=false\n");
    //fprintf(outputfile,"%% fitpage=false\n");
    //fprintf(outputfile,"%% xyratio=0.3\n");
    fprintf(outputfile,"%% ymin=0\n %% ymax=1.1\n");
    fprintf(outputfile,"\n");
    fprintf(outputfile,"%% linelabel=\"measurement data\"\n");
    fprintf(outputfile,"%% linecolor=%d\n",1);
    fprintf(outputfile,"%% linetype =%d\n",1);
    fprintf(outputfile,"%% markertype =%d\n",10);
    for(i=0;i<mpoints;i++) {
      fprintf(outputfile,"%f %f\n",measure[i][0][0], measure[i][1][0]);
    }
    fprintf(outputfile,"\n\n");
    fprintf(outputfile,"%% linelabel=\"calculated data\"\n");
    fprintf(outputfile,"%% linecolor=%d\n",2);
    fprintf(outputfile,"%% linetype =%d\n",2);
    for(i=0;i<xi_max;i++) {
      //fprintf(outputfile,"%f %f\n",xy[i][0][0],
      //  cupmix[1+(int)(t/cupmix_intervall)][i]);
      fprintf(outputfile,"%f %f\n",xy[i][0][0],
        cupmix[1+times][i]);

    }
    fprintf(outputfile,"\n\n");
    fprintf(outputfile,"$ END\n");
    fclose(outputfile);
  }
}

void multiplerunsplot()
{
  int i,n;
  FILE *multRunPlotfile;
  n=times;
  multRunPlotfile=fopen(multiplerunsplotfilename,"w");
  if (multRunPlotfile==NULL) {
    /* opening was unsuccessful */
    fprintf(stderr,"Couldn't open %s for write access .\n", multiplerunsplotfilename);
    exit(8);
  }
  else {
    printf ("Saving multiple runs plot file.\n");
    fprintf(multRunPlotfile, "$ DATA=CURVE2D\n");
    fprintf(multRunPlotfile, "%% toplabel = \"Cup-Mixing Average Chlorine");
    fprintf(multRunPlotfile, " Concentration with time\"\n");
    fprintf(multRunPlotfile, "%% xlabel = \"X\"\n");
    fprintf(multRunPlotfile, "%% ylabel = \"C\"\n");
    fprintf(multRunPlotfile, "%% ymin=0 ymax=1.1\n");
    fprintf(multRunPlotfile, "%% linelabel=\"T = %f\"\n",t);
    fprintf(multRunPlotfile, "%% linecolor=%d\n",0);
    fprintf(multRunPlotfile, "%% linetype =%d\n",runNumber);
    for (i=0;i<xi_max;i++)
      fprintf(multRunPlotfile," %f %f\n",xy[i][1][0],cupmix[n][i]);
    fprintf(multRunPlotfile, "$ END\n");
    fclose(multRunPlotfile);
  }
}

void locationplot()
{
  int i,n;
  double conc;
  FILE *locPlotFile;
  n=times;
  locPlotFile=fopen(locationwithtimeplotfilename,"a");
  if (ftell(locPlotFile)<=0) {
    /* file does not exist, we need to create it and the header */
    //fclose(locPlotFile);
    //locPlotFile=fopen(locationwithtimeplotfilename,"w");
    fprintf(locPlotFile, "$ DATA=CURVE2D\n");
    fprintf(locPlotFile, "%% toplabel = \"Cup-Mixing Average Chlorine");
    fprintf(locPlotFile, " Concentration with time for discrete locations along the pipe\"\n");
    fprintf(locPlotFile, "%% xlabel = \"t\"\n");
    fprintf(locPlotFile, "%% ylabel = \"C\"\n");
    fprintf(locPlotFile, "%% ymin=0 ymax=1.1\n");
    fprintf(locPlotFile, "%% linelabel=\"%s\"\n",plotlabel);
    fprintf(locPlotFile, "%% linecolor=%d\n",0);
  }
  fclose(locPlotFile);
  locPlotFile=fopen(locationwithtimeplotfilename,"a");
  if (locPlotFile==NULL) {
    /* opening for append was unsuccessful */
    fprintf(stderr,"Couldn't open %s for write access .\n", multiplerunsplotfilename);
    exit(8);
  }
  else {
    for(i=1;i<xi_max;i++)
      if ((xy[i][1][0]>=location)&&(xy[i-1][1][0]<location)) conc=cupmix[n][i];
    fprintf(locPlotFile," %f %f\n",locPlot_time, conc);
    fclose(locPlotFile);
  }
}

void exportaverage()
{
  int i,j,n;
  double conc;
  double dx1,dx2;
  //n=1+(int)(t/cupmix_intervall);
  n=times;
  printf("%d\n",n);
  printf("Exported average concentrations: ");
  for(i=1;i<xi_max;i++)
    for(j=0;j<Nr_cup_user;j++)
      if ((xy[i][1][0]>=Xcup_user[j])&&(xy[i-1][1][0]<Xcup_user[j])) {
        dx1 = xy[i][1][0]-xy[i-1][1][0];
        dx2 = Xcup_user[j]-xy[i-1][1][0];
        conc = cupmix[n][i-1]+(cupmix[n][i]-cupmix[n][i-1])*dx2/dx1;
        printf("%e ",conc);
        //printf("%e ",cupmix[n][i]);
      }
  printf("\n");
}

void readconfigfile()
{
  char opt[255];
  char *optarg;
  FILE *configfile;
  char *dummy;
  /* initialize variables */
  optarg=malloc(255);
  dummy=malloc(255);
  if ( (dummy==NULL)||(optarg==NULL)) {
    fprintf(stderr,"No memory available!!\n");
    exit(1);
  }
  else {
    /* attempt to open file */
    configfile = fopen(configfilename,"r");
    /* test result */
    if (configfile==NULL) {
      /* opening was unsuccessful */
      fprintf(stderr,"Couldn't open %s for read access.\n", configfilename);
      exit(8);
    }
    else {
      /* opening successful */
      /* read each line */
      while (fscanf(configfile,"%s\n", opt)!=EOF) {
        /* extract parameter */
        strcpy(optarg,strstr(opt,"=")+1);
        /* determine keyword in line and store parameter in
           corresponding variable */
        if (strstr(opt,"[INPUTFILE]=")==opt) strcpy(inputfilename,optarg);
        if (strstr(opt,"[MEASUREMENTFILE]=")==opt) strcpy(measurementfilename,optarg);
        if (strstr(opt,"[MEASUREMENTPLOTFILE]=")==opt) strcpy(measureplotfilename,optarg);
        if (strstr(opt,"[OUTPUTFILE]=")==opt) strcpy(outputfilename,optarg);
        if (strstr(opt,"[TECPLOTFILE]=")==opt) strcpy(tecplotfilename,optarg);
        if (strstr(opt,"[CUPMIXFILE]=")==opt) strcpy(cupmixfilename,optarg);
        if (strstr(opt,"[ITERATIONS]=")==opt) {
          strcpy(dummy,optarg);
          iterations=atol(dummy);
        }
        if (strstr(opt,"[A0A]=")==opt) {
          strcpy(dummy,optarg);
          A0a=atof(dummy);
        }
        if (strstr(opt,"[A0R]=")==opt) {
          strcpy(dummy,optarg);
          A0r=atof(dummy);
        }
        if (strstr(opt,"[A11N]=")==opt) {
          strcpy(dummy,optarg);
          A11n=atof(dummy);
        }
        if (strstr(opt,"[A12N]=")==opt) {
          strcpy(dummy,optarg);
          A12n=atof(dummy);
        }
        if (strstr(opt,"[A11M]=")==opt) {
          strcpy(dummy,optarg);
          A11m=atof(dummy);
        }
        if (strstr(opt,"[A12M]=")==opt) {
          strcpy(dummy,optarg);
          A12m=atof(dummy);
        }
        if (strstr(opt,"[A2]=")==opt) {
          strcpy(dummy,optarg);
          A2=atof(dummy);
        }
        if (strstr(opt,"[FLOW]=")==opt) {
          if (strstr(optarg,"turbulent")==optarg)
            flow=1;
          else if (strstr(optarg,"laminar")==optarg)
              flow=0;
            else flow = 2; // stagnation
        }
        if (strstr(opt,"[AVERAGE]=")==opt) {
          if (strstr(optarg,"cupmix")==optarg)
            average=1; // calculate cupmixing average
          else average=0; // calculate area averaged concentration
        }
        if (strstr(opt,"[VELOCITY]=")==opt) {
          strcpy(dummy,optarg);
          if (atoi(dummy)) velocityprofile=1;
        }
        if (strstr(opt,"[TIMESTEP]=")==opt) {
          strcpy(dummy,optarg);
          d_t=atof(dummy);
        }
        if (strstr(opt,"[CUPMIXIV]=")==opt) {
          strcpy(dummy,optarg);
          cupmix_intervall=atoi(dummy);
        }
        if (strstr(opt,"[C01]=")==opt) {
          strcpy(dummy,optarg);
          c01=atof(dummy);
        }
        if (strstr(opt,"[C02]=")==opt) {
          strcpy(dummy,optarg);
          c02=atof(dummy);
        }
        if (strstr(opt,"[RE]=")==opt) {
          strcpy(dummy,optarg);
          Re=atof(dummy);
        }
        if (strstr(opt,"[R0]=")==opt) {
          strcpy(dummy,optarg);
          r0=atof(dummy);
        }
        if (strstr(opt,"[A0AE]=")==opt) {
          strcpy(dummy,optarg);
          A0ae=atof(dummy);
        }
        if (strstr(opt,"[A0RE]=")==opt) {
          strcpy(dummy,optarg);
          A0re=atof(dummy);
        }
        //
        if (strstr(opt,"[PLOTLABEL]=")==opt) strcpy(plotlabel,optarg);
        if (strstr(opt,"[MULTIPLERUNSPLOTFILENAME]=")==opt) strcpy(multiplerunsplotfilename,optarg);
        if (strstr(opt,"[RUN]=")==opt) {
          strcpy(dummy,optarg);
          runNumber=atof(dummy);
        }
        //
        if (strstr(opt,"[LOCATIONWITHTIMEPLOTFILENAME]=")==opt) strcpy(locationwithtimeplotfilename,optarg);
        if (strstr(opt,"[LOCATION]=")==opt) {
          strcpy(dummy,optarg);
          location=atof(dummy);
        }
        if (strstr(opt,"[TIME]=")==opt) {
          strcpy(dummy,optarg);
          locPlot_time=atof(dummy);
        }
        //
        if (strstr(opt,"[EXPORTAVERAGEAT]=")==opt) {
          strcpy(dummy,optarg);
          Xcup_user[Nr_cup_user]=atof(dummy);
          if (Nr_cup_user<19) Nr_cup_user++;
        }
        // average velocity
        if ((r0>0)&&(Re>0)&&(flow==1)) {
          u0=Re*viscosity/r0/2;
          printf("average velocity: %e m/s\n", u0);
          printf("pipe radius: %e m\n",r0);
          printf("A0ae: %e \n",A0ae);
          printf("A0re: %e \n",A0re);
        }
      }
      /* close file */
      fclose(configfile);
    }
    free(optarg);
    free(dummy);
  }
}

/* CHECKINPUT - prints and tests all input parameters and decides if the pro-
    gram can be run with the given set of parameters                       */

void checkinput()
{
  /* inputfile available */
  is_inputfile=1;
  if (inputfilename=='\0'){
    fprintf(stderr,"Input file name not provided, execution will be stopped!\n"); is_inputfile=0;
    exit(8);
  } else printf("Input file:        %s\n",inputfilename);
  /* timestep */
  if (d_t<=0) {
    fprintf(stderr,"Incorrect value for timestep used, execution will be stopped!\n");
    exit(8);
  } else printf("Timestep:          %f\n",d_t);
  if (iterations<=0) {
    fprintf(stderr,"Specified nomber of iterations is wrong, execution will be stopped!\n");
    exit(8);
  } else printf("Iterations:        %ld\n",iterations);
  /* outputfile available */
  is_outputfile=1;
  if (strlen(outputfilename)==0){
    fprintf(stderr,"Output file name not provided, results will not be stored!\n");
    is_outputfile=0;
  } else printf("Output file:       %s\n",outputfilename);
  is_tecplotfile=1;
  if (strlen(tecplotfilename)==0){
    fprintf(stderr,"Tecplot file name not provided, no tecplot file will be created.\n");
    is_tecplotfile=0;
  } else printf("TecPlot file:      %s\n",tecplotfilename);
  is_measure=1;
  if (strlen(measurementfilename)==0){
    fprintf(stderr,"Measurement file name not provided, results will not be compared to measured results\n");
    is_measure=0;
  } else printf("Measurement file:       %s\n",measurementfilename);
  is_cupmix=1;
  if (strlen(cupmixfilename)==0){
    fprintf(stderr,"Name of cupmixing average file not provided,\n");
    fprintf(stderr," file will not be created.\n");
    is_cupmix=0;
  } else {
    printf("Cupmixing avg.file:%s\n",cupmixfilename);
    if (average==0) printf("Cross-sectional averaging\n");
    else printf("Cup-mixing averaging\n");
  }
  if (cupmix_intervall<=0) {
    fprintf(stderr,"Cupmixing average calcuation intervall not provided,");
    fprintf(stderr," no calculation of");
    fprintf(stderr," cupmixing average.\n");
    is_cupmix=0;
  } else{
    if (cupmix_intervall>=iterations){
      cupmix_intervall=iterations;
    }
    printf("Cupmix. intervall: %d\n",cupmix_intervall);
  }
  if (c02==0)
    dc0=0;
  else if (iterations>0){
      dc0=(c02-c01)/((float)iterations+1);
      printf("C01 = %f\n",c01);
      printf("C02 = %f\n",c02);
      printf("dc0 = %f\n",dc0);
  }
  is_multiplerunsplot=1;
  if ((strlen(multiplerunsplotfilename)==0)||(runNumber<=0)) {
    is_multiplerunsplot=0;
    fprintf(stderr, "Multiple runs series file will not be stored\n");
  }
  is_locationplot=1;
  if ((location<0)||(location>1)||(strlen(locationwithtimeplotfilename)==0)
       ||(locPlot_time<0)) {
    is_locationplot=0;
    fprintf(stderr, "Concentration at a location will not be stored\n");
  }
}


/* MATRIX_SOLVER_TEST - execute a test run to see if the matrix solver
    (functions at top) really works as it should..                         */

void matrix_solver_test()
{
  int i;
  A[0][0][0]=0; A[1][0][0]=1, A[2][0][0]=1; A[3][0][0]=6;
  B[0][0][0]=1; B[1][0][0]=4, B[2][0][0]=2; B[3][0][0]=8;
  C[0][0][0]=2; C[1][0][0]=5, C[2][0][0]=1; C[3][0][0]=0;
  RHS[0][0]=5; RHS[1][0]=14; RHS[2][0]=5; RHS[3][0]=14;
  block_tridiagonal(4,A,B,C,RHS);
  for(i=0;i<4;i++){
    fprintf(stdout,"%f\n",RHS[i][0]);
  }
}

/***************************************************************************/
/* MAIN - main program loop                                                */

int main(int argc, char *argv[])
{
  extern char *optarg;
  extern int optind, optopt;
  char *commandline;
  int i;
  /********************************************************************/
  /* initialize variables */
  inputfilename = malloc(255);
  measurementfilename = malloc(255);
  outputfilename = malloc(255);
  tecplotfilename = malloc(255);
  cupmixfilename = malloc(255);
  commandline=malloc(255);
  configfilename=malloc(255);
  measureplotfilename=malloc(255);
  plotlabel=malloc(255);
  multiplerunsplotfilename=malloc(255);
  locationwithtimeplotfilename=malloc(255);
  if ((configfilename==NULL)||
      (inputfilename==NULL)||
      (measurementfilename==NULL)||
      (outputfilename==NULL)||
      (tecplotfilename==NULL)||
      (measureplotfilename==NULL)||
      (cupmixfilename==NULL)||(commandline==NULL)||
      (multiplerunsplotfilename==NULL)||(plotlabel==NULL)||
      (locationwithtimeplotfilename==NULL)){
    printf("Not enough memory.\nExiting\n");
    exit(8);
  }
  /* Inlet concentrations */
  c01=1; c02=0;
  /* 1. Load Configuration */
  /* extract path from commandline*/
  /* (maybe put this in its own subroutine?) */
  strcpy(commandline,argv[0]);
  i=strlen(commandline);
  /* Check if command line contains a slash ("/"), counting backwards
  from the end of the string */
  do { i--; } while ((commandline[i]!='/') && (i>1));
  /* slash found. The part before the slash should be the
  current directory -> extract it. */
  if (i>1) strncpy(configfilename,commandline,i+1);
  else {
    /* no backslash found. Maybe we run under Windows? If this is the case,
    we must look for a backslash ("\") in the command line */
    i=strlen(commandline); /* check from the end of the string ..*/
    do { i--; } while ((commandline[i]!='\\') && (i>0)); /* look for "\" */
    /* backslash found, extract string before it as current directory */
    if (i>1) strncpy(configfilename,commandline,i+1);
    else {
      /* no backslash found! OK, then we assume that the current path
      is an empty string, but we warn the user about it */
      printf("Current path could not be determined! Trying to");
      printf("proceed without it...\n");
      strcpy(configfilename,"");
    }
  }
  /* path to solver.rc is combination of current path and file name */
  strcat(configfilename,"solver.rc");
  printf("Configuration file: %s\n",configfilename);
  /* load */
  readconfigfile();
  /* check for missing information */
  checkinput();
  /* print information */
  printf("A0,a:              %e\n",A0a);
  printf("A0,r:              %e\n",A0r);
  printf("A1,1m*x+A1,1n:     %f*x+%f\n",A11m, A11n);
  printf("A1,2m*x+A1,2n:     %f*x+%f\n",A12m, A12n);
  printf("A2:                %f\n",A2);
  if (flow ==1)
    printf("turbulent flow\n");
  else if (flow==0 )printf("laminar flow\n");
  else printf("stagnant flow\n");
  /************************************************************************/
  /* 2. Read the input file */
  readinputfile();
  // logfile=fopen(logfilename,"w");
  /************************************************************************/
  /* 3. Read the file containing measurement data */
  /************************************************************************/
  if (is_measure) readmeasurementfile();
  /* 4. Start iteration */
  if (is_cupmix) open_cupmixfile_mtv();
  solve();
 //matrix_solver_test();
  /************************************************************************/
  /* 5. Save results */
  /*   dimensionalizexy(); */
  if (is_cupmix) close_cupmixfile_mtv();
  if (is_measure) extract_data();
  if (is_outputfile) savefile();
   /* if (normalized==1) dimensionalize(); */
  // if (is_tecplotfile) outputtecplot();
  if (is_tecplotfile) outputmtv();
  if (is_measure) measureplot();
  if (is_multiplerunsplot) multiplerunsplot();
  if (Nr_cup_user>0) exportaverage();
  if (is_locationplot) locationplot();
  // fclose(logfile);
  //printf("Re = %5.1f Cout = %e\n",Re,cupmix[0][xi_max-1]);
  //printf("Concentrations: 0: %e 5: %e 9: %e\n",Ccup[0],Ccup[5],Ccup[9]);
  free(commandline);
  free(configfilename);
  free(inputfilename);
  free(measurementfilename);
  free(outputfilename);
  free(tecplotfilename);
  free(cupmixfilename);
  free(measureplotfilename);
  free(plotlabel);
  free(multiplerunsplotfilename);
  free(locationwithtimeplotfilename);
  return (0);
}

